'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var Global = _interopDefault(require('nexie'));

const COMPONENT_CACHE = /*#__PURE__*/new Map();
const DEFAULT_SHADOWMODE = {
  mode: 'open'
};
const COMPONENTS = /*#__PURE__*/Global.set( /*#__PURE__*/Symbol.for('@wayang/web-components'), /*#__PURE__*/new Map());
const PRESENTERS = /*#__PURE__*/Global.set( /*#__PURE__*/Symbol.for('@wayang/presenters'), /*#__PURE__*/new Map());
const MIXINS = /*#__PURE__*/Global.set( /*#__PURE__*/Symbol.for('@wayang/mixins'), /*#__PURE__*/new Map());
const CONVERTERS = /*#__PURE__*/new Map([['string', value => {
  return value;
}], ['integer', value => {
  return parseInt(value);
}], ['float', value => {
  return parseFloat(value);
}], ['number', value => {
  return Number(value);
}], ['json', value => {
  return JSON.parse(value);
}], ['boolean', value => {
  return value.toLowerCase() === 'true';
}], ['date', value => {
  return new Date(value);
}], ['ticks', value => {
  return new Date(Number(value));
}], ['exists', value => {
  return value !== null;
}]]);
const NO_TAG = '!no-tag';
const NO_HTML = '!no-html';
const NO_CSS = '!no-css';

class CacheItem {
  constructor(tag, template, styles, component, converters, properties, observed) {
    this.mixins = new Map();
    this.tag = tag;
    this.template = template;
    this.styles = styles;
    this.component = component;
    this.converters = new Map(converters);
    this.properties = new Map(properties);
    this.observed = observed.slice(0);
  }

}

const instanceKey = /*#__PURE__*/Symbol.for(`@wayang/presenter/instance`);

class Presenter {
  constructor() {
    this.global = Global;
  }

  static register() {
    // basic validation
    if (!this.id) {
      throw new Error('No id was defined for this presenter');
    }

    if (PRESENTERS.has(this.id)) {
      throw new Error(`The presenter id ${this.id} has already been registered`);
    } // registration


    PRESENTERS.set(this.id, this);
    Global.set(Symbol.for(`@wayang/presenters/${this.id}`), this);
  }

  static new() {
    return new this();
  }

  static addTo(component, id) {
    Global.request(Symbol.for(`@wayang/presenters/${id}`)).then(presenter => {
      if (component[instanceKey]) {
        return;
      }

      component[instanceKey] = presenter.new();
      component[instanceKey].connect(component);
    });
  }

  static removeFrom(component) {
    if (!component[instanceKey]) {
      return;
    }

    component[instanceKey].disconnect(component);
    component[instanceKey] = null;
  }

}

const theme = '@wayang/theme';

class Style {
  constructor() {
    const id = this.getId();
    const key = `@wayang/style/${id}`;
    const exists = Global.has(key);

    if (exists) {
      throw new Error(`A style with an id of '${id}' already exists.`);
    }

    Global.set(key, this);
  }

  static get theme() {
    return Global.get(theme);
  }

  static set theme(name) {
    Global.upsert(theme, name);
  }

  getId() {
    return this.id;
  }

}
// class ButtonTheme extends Style {
//     public readonly id = 'button-theme';
//     public readonly prefetch = true;
//     public async content(): Promise<string> {
//         const request = await fetch(`./button/${ Style.theme }`);
//         return request.text();
//     }
// }

class WebComponent extends HTMLElement {
  constructor() {
    super();
    this.global = Global;
    this.debouncers = {};
    this.class = this.constructor;
    this.root = this.attachShadow(this.class.mode);
    this.root.appendChild(this.class.cache.template.content.cloneNode(true));
    this.root.appendChild(this.class.cache.styles.content.cloneNode(true));
    this.class.mixins.forEach(mixin => mixin.connect(this));
  } // required shadow element members


  static get observedAttributes() {
    return this.cache.observed;
  }

  connectedCallback() {
    this.dispatch('connected', this);
  }

  adoptedCallback() {
    this.dispatch('adopted', this);
  }

  disconnectedCallback() {
    this.dispatch('disconnected', this);
    this.destructor();
  }

  attributeChangedCallback(name, previous, current) {
    if (previous === current) {
      return;
    }

    clearTimeout(this.debouncers[name]);
    this.debouncers[name] = setTimeout(() => {
      const properties = this.class.cache.properties;
      const converters = this.class.cache.converters;

      const converter = converters.get(properties.get(name) || 'string') || (v => v);

      const _old = this[name];

      const _new = converter(current);

      this[name] = _new;
      this.dispatch(`${name}Changed`, {
        old: _old,
        new: _new
      });
    }, 0);
  } // element instance properties


  get presenter() {
    return this.getAttribute('presenter') || '';
  }

  set presenter(value) {
    value ? this.setAttribute('presenter', value) : this.removeAttribute('presenter');
    Presenter.removeFrom(this);
    Presenter.addTo(this, value);
  } // element instance methods


  dispatch(eventName, data) {
    return this.dispatchEvent(new CustomEvent(eventName, {
      detail: data,
      bubbles: true,
      cancelable: true
    }));
  }

  listen(eventName, listener) {
    this.addEventListener(eventName, listener);
    return this;
  }

  unlisten(eventName, listener) {
    this.removeEventListener(eventName, listener);
    return this;
  }

  select(selector) {
    return Array.prototype.slice.call(this.root.querySelectorAll(selector));
  } // special members


  static get cache() {
    return COMPONENT_CACHE.get(this.tag);
  }

  static register() {
    // basic validation
    if (this.tag === NO_TAG) {
      throw new Error('static tag: string, must be redefined.\nString must be a valid custom element tag.');
    }

    if (this.html === NO_HTML) {
      throw new Error('static html: string, must be redefined.');
    }

    if (this.css === NO_CSS) {
      throw new Error('static css: string, must be redefined.');
    } // register


    if (this.cache) {
      return;
    } // template


    const template = document.createElement('template');
    template.innerHTML = this.html; // styles

    const styles = document.createElement('template');
    let css = this.css;

    if (!Array.isArray(css)) {
      css = [css];
    }

    styles.innerHTML = css.map(style => {
      if (style instanceof Style) {
        return `<style id="style-${style.id}" type="text/css"></style>`;
      }

      return `<style type="text/css">\n${style}\n</style>`;
    }).join('\n'); // observed props

    const observed = this.observed; // built-in props

    observed.set('presenter', 'string');
    const properties = [...observed.keys()];
    const cache = new CacheItem(this.tag, template, styles, this, this.converters, this.observed, properties);
    COMPONENT_CACHE.set(this.tag, cache);
    COMPONENTS.set(this.tag, this);
    this.mixins.forEach(mixin => mixin.attach(this));
    customElements.define(this.tag, this);
    Global.set(Symbol.for(`@wayang/web-components/${this.tag}`), this);
  }

  destructor() {
    Presenter.removeFrom(this);
    this.class.cache.mixins.forEach(mixin => {
      mixin.disconnect(this);
      const mixinKey = Symbol.for(`@wayang/mixin/state/${mixin.id}`);
      this[mixinKey] = null;
      delete this[mixinKey];
    });
  }

} // must be re-defined


WebComponent.tag = NO_TAG;
WebComponent.html = NO_HTML;
WebComponent.css = NO_CSS;
WebComponent.observed = /*#__PURE__*/new Map(); // optionally re-defined

WebComponent.converters = /*#__PURE__*/new Map(CONVERTERS);
WebComponent.mode = DEFAULT_SHADOWMODE;
WebComponent.mixins = /*#__PURE__*/new Array();

class Mixin {
  static connect(component) {
    throw new Error('static connect(component: WebComponent), must be redefined');
  }

  static disconnect(component) {
    throw new Error('static disconnect(component: WebComponent), must be redefined');
  } // must re-define


  static html(html) {
    throw new Error('static html(html: HTMLTemplateElement): void, must be redefined');
  }

  static css(css) {
    throw new Error('static css(css: HTMLTemplateElement): void, must be redefined');
  }

  static properties(properties) {
    throw new Error('static properties(properties: Map<string, string>): void, must be redefined');
  }

  static observed(observed) {
    throw new Error('static observed(observed: string[]): void, must be redefined');
  }

  static converters(converters) {
    throw new Error('static converters(converters: Map<string, (value: string) => any>): void, must be redefined');
  } // special members


  static register() {
    const id = this.id;

    if (!id) {
      throw new Error('No id was defined for this mixin');
    }

    if (MIXINS.has(id)) {
      throw new Error(`The mixin id ${id} has already been registered`);
    }

    MIXINS.set(id, this);
  }

  static stateOf(component) {
    const key = `@wayang/mixin/state/${this.id}`;

    if (!component.hasOwnProperty(key)) {
      component[key] = new Map();
    }

    return component[key];
  }

  static attach(component) {
    const cache = component.cache;

    if (cache.mixins.has(this.id)) {
      return;
    }

    cache.mixins.set(this.id, this);

    if (!component.mixins.includes(this)) {
      component.mixins.push(this);
    }

    this.html(cache.template);
    this.css(cache.styles);
    this.properties(cache.properties);
    this.observed(cache.observed);
    this.converters(cache.converters);
  }

}

exports.Mixin = Mixin;
exports.Presenter = Presenter;
exports.WebComponent = WebComponent;
exports.default = WebComponent;
//# sourceMappingURL=wayang-ui.cjs.development.js.map
